from __future__ import annotations

import os
from typing import TYPE_CHECKING

import pytest

from thanosql._error import ThanoSQLNotFoundError, ThanoSQLValueError

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL

dir_name = "test"
file_name = "file_image.jpeg"
relative_file_path = f"{dir_name}/{file_name}"
temp_file_name = "temp"


def test_create_folder(client: ThanoSQL):
    res = client.file.create(dir_name)
    assert res.content_info.name == dir_name
    # we will check whether the folder is created successfully again
    # when we upload a file to it in the next test


def test_upload_file(client: ThanoSQL):
    # nonexistent file in local
    with pytest.raises(ThanoSQLNotFoundError):
        client.file.create(file="random_file")

    res = client.file.create(dir_name, file_name)
    assert res.content_info.path == relative_file_path

    # cannot upload to file as destination
    with pytest.raises(ThanoSQLValueError):
        client.file.create(relative_file_path, file_name)


def test_get_files(client: ThanoSQL):
    # nonexistent file/directory
    with pytest.raises(ThanoSQLNotFoundError):
        client.file.get("random_file")

    res = client.file.get(relative_file_path)
    assert "".join([res.content_info.name, res.content_info.format]) == file_name
    assert res.content_info.path == relative_file_path
    assert res.content_info.type == "file"

    res = client.file.get()
    assert res.content_info.type == "directory"
    for content in res.content_info.content:
        if content["name"] == dir_name:
            return
    raise AssertionError("Test folder not found in root directory")


def test_download_file(client: ThanoSQL):
    # we cannot download a directory
    with pytest.raises(ThanoSQLValueError):
        client.file.get(dir_name, "download")

    # upload a dummy file that we can delete later
    with open(temp_file_name, "w") as f:
        f.write("Hello ThanoSQL!")

    res = client.file.create("/", temp_file_name)
    assert res.content_info.path == temp_file_name

    res = client.file.get(temp_file_name, "download")
    assert os.path.isfile(temp_file_name)

    # cleanup
    os.remove(temp_file_name)


def test_delete_file(client: ThanoSQL):
    client.file.delete(dir_name)
    client.file.delete(temp_file_name)

    # nonexistent file/directory -> we already deleted the folder
    # make sure deletion deletes all folder contents as well
    with pytest.raises(ThanoSQLNotFoundError):
        client.file.delete(relative_file_path)



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from pydantic import Field, TypeAdapter

from thanosql._service import ThanoSQLService
from thanosql.resources import Column
from thanosql.resources._model import BaseModel

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL


class View(BaseModel):
    name: str
    table_schema: Optional[str] = Field(alias="schema", default=None)
    columns: List[Column] = []
    definition: str = ""


class ViewService(ThanoSQLService):
    """Service layer for view methods.

    Attributes
    ----------
    client: ThanoSQL
        The ThanoSQL client used to make requests to the engine.

    """

    def __init__(self, client: ThanoSQL) -> None:
        super().__init__(client=client, tag="view")

    def list(
        self,
        schema: Optional[str] = None,
        verbose: Optional[bool] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List[View]:
        """Lists views stored in the workspace.

        Parameters
        ----------
        schema: str, optional
            When specified, lists views from a specific schema only.
            Otherwise, lists views from all available schemas.
        verbose: bool, optional
            When not set or set to False (default behavior), only lists
            view names and the schema they belong to. When set to True,
            includes column descriptions and view definition in the output
            on top of the basic (name and schema) information.
        offset: int, optional
            When set to n, skips the first n views and excludes them from
            the output list. Otherwise, starts the list from the first view
            stored. Must be greater than 0.
        limit: int, optional
            When set to n, limits the number of views listed to n. Otherwise,
            lists up to 100 views per call. Must range between 0 to 100.

        Returns
        -------
        List[View]
            A list of View objects.

        Raises
        ------
        ThanoSQLValueError
            If offset is less than 0 or if limit is not between 0 to 100 (inclusive).

        """
        path = f"/{self.tag}/"
        query_params = self._create_input_dict(
            schema=schema,
            verbose=verbose,
            offset=offset,
            limit=limit,
        )

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        views_adapter = TypeAdapter(List[View])
        parsed_response = views_adapter.validate_python(raw_response["views"])
        return parsed_response

    def get(self, name: str, schema: Optional[str] = None) -> View:
        """Shows the details of the specified view.

        Parameters
        ----------
        name: str
            The name of the view to be retrieved.
        schema: str, optional
            The schema the target view is stored in. When not specified,
            defaults to "public".

        Returns
        -------
        View
            A View object.

        """
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        view_adapter = TypeAdapter(View)
        parsed_response = view_adapter.validate_python(raw_response["view"])
        return parsed_response

    def delete(self, name: str, schema: Optional[str] = None) -> dict:
        """Deletes the specified view.

        Parameters
        ----------
        name: str
            The name of the view to be deleted.
        schema: str, optional
            The schema the target view is stored in. When not specified,
            defaults to "public".

        Returns
        -------
        dict
            A dictionary containing a success message and the name of the
            deleted view in the format of::

                {
                    "message": "string",
                    "view_name": "string"
                }

        """
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)

        return self.client._request(
            method="delete", path=path, query_params=query_params
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from thanosql._service import ThanoSQLService

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL


class SchemaService(ThanoSQLService):
    """Service layer for schema methods.

    Attributes
    ----------
    client: ThanoSQL
        The ThanoSQL client used to make requests to the engine.

    """

    def __init__(self, client: ThanoSQL) -> None:
        super().__init__(client=client, tag="schema")

    def list(self) -> dict:
        """Lists schemas stored in the workspace.

        Does not have any parameters.

        Returns
        -------
        dict
            A dictionary containing the names of stored schemas in the format of::

                {
                    "schemas": [
                        {
                            "name": "string"
                        }
                    ]
                }

        """
        path = f"/{self.tag}/"

        return self.client._request(method="get", path=path)

    def create(self, name: str) -> dict:
        """Creates a new schema in the workspace.

        Parameters
        ----------
        name: str
            The name of the schema to be created.

        Returns
        -------
        dict
            A dictionary containing the name of the created schema and
            a success message in the format of::

                {
                    "schema": "string",
                    "message": "string"
                }

        """
        path = f"/{self.tag}/{name}"

        return self.client._request(method="post", path=path)



from __future__ import annotations

import enum
from datetime import datetime
from typing import TYPE_CHECKING, List, Optional

from pydantic import TypeAdapter

from thanosql._error import ThanoSQLValueError
from thanosql._service import ThanoSQLService
from thanosql.resources._model import BaseModel
from thanosql.resources._record import Records

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL


class QueryLog(BaseModel):
    query_id: Optional[str]
    statement_type: Optional[str]
    start_time: Optional[datetime]
    end_time: Optional[datetime]
    query: str
    referer: str
    state: Optional[str]
    destination_table_name: Optional[str]
    destination_schema: Optional[str]
    error_result: Optional[str]
    created_at: Optional[datetime]
    records: Optional[Records] = None


class QueryType(enum.Enum):
    THANOSQL = "thanosql"
    PSQL = "psql"


class QueryService(ThanoSQLService):
    """Service layer for query methods.

    Attributes
    ----------
    client: ThanoSQL
        The ThanoSQL client used to make requests to the engine.
    log: QueryLogService
        The query log service layer to access methods involving query logs.
    template: QueryTemplateService
        The query template service layer to access methods involving
        query templates.

    """

    def __init__(self, client: ThanoSQL) -> None:
        super().__init__(client=client, tag="query")

        self.log: QueryLogService = QueryLogService(self)
        self.template: QueryTemplateService = QueryTemplateService(self)

    def execute(
        self,
        query: Optional[str] = None,
        query_type: str = "thanosql",
        template_id: Optional[int] = None,
        template_name: Optional[str] = None,
        parameters: Optional[dict] = None,
        schema: Optional[str] = None,
        table_name: Optional[str] = None,
        overwrite: Optional[bool] = None,
        max_results: int = 100,
    ) -> QueryLog:
        """Executes a query string.

        There are three ways of requesting a query:

        - Using a complete query directly in `query`, leaving
          `template_id`, `template_name`, and `parameters` empty.
        - Using a template query in `query` and completing it with `parameters`,
          leaving `template_id` and `template_name` empty.
        - Recalling a template query from the database using `template_id` or
          `template_name` (but not both) and completing it with `parameters`,
          leaving `query` empty.

        One, and only one, of these ways must be chosen. That is, exactly one of
        `query`, `template_id`, or `template_name` must be specified. If none or
        more than one way is selected, an error will occur.

        Parameters
        ----------
        query: str, optional
            The query string or template to be executed.
        query_type: str, default "thanosql"
            The type of the query to be executed. Can only be one of
            "thanosql" or "psql".
        template_id: int, optional
            The ID number of the query template to be used.
            Only relevant when a query template from the database is needed,
            and `query` and `template_name` are not used simultaneously.
        template_name: str, optional
            The name of the query template to be used.
            Only relevant when a query template from the database is needed,
            and `query` and `template_id` are not used simultaneously.
        parameters: dict, optional
            A dictionary of parameter names and values to fill in the template.
            Only relevant when a query template, either from `query` or the database,
            is used.
        schema: str, optional
            The schema of the table to save the query results in. If not specified
            and table_name is also empty, it defaults to "qm". If table_name is
            specified, it defaults to "public".
        table_name: str, optional
            The name of the table to save the query results in. If not specified
            while the query produces a result, an automatic table name will be given.
        overwrite: bool, optional
            Whether to overwrite the table if a table with the same `table_name`
            and `schema` already exists. If not specified, the value is False.
        max_results: int, optional
            The maximum number of records to be returned by the response QueryLog.
            If not specified, it defaults to 100.

        Returns
        -------
        QueryLog
            A query log object containing details about the results of the
            executed query.

        Raises
        ------
        ThanoSQLValueError
            - If invalid input combination is provided; that is:
                - query, template_id, and template_name are all empty, or
                - more than one of query, template_id, and template_name are non-empty
            - If a value other than "thanosql" or "psql" is provided as query_type.
            - If max_results is not between 0 and 100 (inclusive).
            - If query and parameters are used but the template has invalid format.
            - If rendering query template by substituting in parameters fails, either \
                by direct query template or templates from the database.

        """
        try:
            query_type_enum = QueryType(query_type)
        except Exception as e:
            raise ThanoSQLValueError(str(e))

        path = f"/{self.tag}/"
        query_params = self._create_input_dict(
            schema=schema,
            table_name=table_name,
            overwrite=overwrite,
            max_results=max_results,
        )
        payload = self._create_input_dict(
            query_type=query_type_enum.value,
            query_string=query,
            template_id=template_id,
            template_name=template_name,
            parameters=parameters,
        )

        raw_response = self.client._request(
            method="post", path=path, query_params=query_params, payload=payload
        )

        query_log_adapter = TypeAdapter(QueryLog)
        parsed_response = query_log_adapter.validate_python(raw_response)
        return parsed_response


class QueryLogService(ThanoSQLService):
    """Service layer for query log methods.

    Cannot exist without a parent QueryService.

    Attributes
    ----------
    query: QueryService
        The parent QueryService to connect to the ThanoSQL client.

    """

    def __init__(self, query: QueryService) -> None:
        super().__init__(client=query.client, tag="log")

        self.query: QueryService = query

    def list(
        self,
        search: Optional[str] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> dict:
        """Lists the details of stored query logs.

        Parameters
        ----------
        search : str, optional
            Search keywords that the query strings in the results must contain.
            If not set, all query logs are returned by default.
        offset : int, optional
            When set to n, skips the first n results and excludes them from
            the output list. Otherwise, starts the list from the first result
            stored. Must be greater than 0.
        limit : int, optional
            When set to n, limits the number of results listed to n. Otherwise,
            lists up to 100 results per call. Must range between 0 to 100.

        Returns
        -------
        dict
            A dictionary of results in the format of::

                {
                    "query_logs": ["QueryLog"],
                    "total": 0
                }

        Raises
        ------
        ThanoSQLValueError
            If offset is less than 0 or if limit is not between 0 to 100 (inclusive).

        """
        path = f"/{self.query.tag}/{self.tag}"
        query_params = self._create_input_dict(
            search=search, offset=offset, limit=limit
        )

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        query_logs_adapter = TypeAdapter(List[QueryLog])
        parsed_response = {}
        parsed_response["query_logs"] = query_logs_adapter.validate_python(
            raw_response["query_logs"]
        )
        parsed_response["total"] = raw_response["total"]

        return parsed_response


class QueryTemplate(BaseModel):
    id: Optional[int] = None
    name: str
    query: str
    parameters: Optional[List[str]] = []
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


class QueryTemplateService(ThanoSQLService):
    """Service layer for query template methods.

    Cannot exist without a parent QueryService.

    Attributes
    ----------
    query: QueryService
        The parent QueryService to connect to the ThanoSQL client.

    """

    def __init__(self, query: QueryService) -> None:
        super().__init__(client=query.client, tag="template")

        self.query: QueryService = query

    def _parse_query_template_response(self, raw_response: dict):
        query_template_adapter = TypeAdapter(QueryTemplate)
        parsed_response = query_template_adapter.validate_python(
            raw_response["query_template"]
        )
        return parsed_response

    def list(
        self,
        search: Optional[str] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        order_by: Optional[str] = None,
    ) -> List[QueryTemplate]:
        """Lists query templates stored in the workspace.

        Parameters
        ----------
        search : str, optional
            Search keywords that the query strings in the results must contain.
            If not set, all query templates are returned by default.
        offset : int, optional
            When set to n, skips the first n results and excludes them from
            the output list. Otherwise, starts the list from the first result
            stored. Must be greater than 0.
        limit : int, optional
            When set to n, limits the number of results listed to n. Otherwise,
            lists up to 100 results per call. Must range between 0 to 100.
        order_by : str, optional
            How to order the results. There are only three possible values:
            - recent: based on the date of creation, from most recent to oldest
            - name_asc: based on the name of the template, from A to Z
            - name_desc: based on the name of the template, from Z to A

        Returns
        -------
        List[QueryTemplate]
            A list of QueryTemplate objects.

        Raises
        ------
        ThanoSQLValueError
            - If offset is less than 0 or if limit is not between 0 to 100 (inclusive).
            - If order_by is not one of "recent", "name_asc", or "name_desc".

        """
        path = f"/{self.query.tag}/{self.tag}"
        query_params = self._create_input_dict(
            search=search, offset=offset, limit=limit, order_by=order_by
        )

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        query_templates_adapter = TypeAdapter(List[QueryTemplate])
        parsed_response = query_templates_adapter.validate_python(
            raw_response["query_templates"]
        )
        return parsed_response

    def create(
        self,
        name: Optional[str] = None,
        query: Optional[str] = None,
        dry_run: Optional[bool] = None,
    ) -> QueryTemplate:
        """Creates a new query template.

        Parameters
        ----------
        name : str, optional
            The name of the query template to be created. If left empty, an
            automatic name will be given.
        query : str, optional
            The string contents of the template.
        dry_run : bool, optional
            Whether to "dry run" the template creation or save it to database.
            When set to True, only shows whether the query template is valid
            or not without actually storing it. By default, created query
            templates will be saved to the workspace database.

        Returns
        -------
        QueryTemplate
            QueryTemplate object of the created query template.

        Raises
        ------
        ThanoSQLValueError
            - If the template name contains invalid characters or is too long.
            - If the query template contains invalid formatting.

        """
        path = f"/{self.query.tag}/{self.tag}"
        query_params = self._create_input_dict(dry_run=dry_run)
        payload = self._create_input_dict(name=name, query=query)

        raw_response = self.client._request(
            method="post", path=path, query_params=query_params, payload=payload
        )

        return self._parse_query_template_response(raw_response)

    def get(self, name: str) -> QueryTemplate:
        """Shows the details of the specified query template.

        Parameters
        ----------
        name : str
            The name of the query template to be retrieved.

        Returns
        -------
        QueryTemplate
            A QueryTemplate object.

        """
        path = f"/{self.query.tag}/{self.tag}/{name}"
        raw_response = self.client._request(method="get", path=path)
        return self._parse_query_template_response(raw_response)

    def update(
        self,
        current_name: str,
        new_name: Optional[str] = None,
        query: Optional[str] = None,
    ) -> QueryTemplate:
        """Updates the specified query template.

        Parameters
        ----------
        current_name : str
            The current name of the query template to be updated.
        new_name : str, optional
            The new name of query template after update. If not set,
            the name will not be changed.
        query : str, optional
            The query contents of the query template after update. If
            not set, the query string will not be changed.

        Returns
        -------
        QueryTemplate
            QueryTemplate object of the new query template after update.

        Raises
        ------
        ThanoSQLValueError
            - If the new template name is set but is empty or null, contains invalid \
                characters, or is too long.
            - If the new query template is set but is null or contains invalid formatting.

        """
        path = f"/{self.query.tag}/{self.tag}/{current_name}"
        payload = self._create_input_dict(name=new_name, query=query)

        raw_response = self.client._request(method="put", path=path, payload=payload)

        return self._parse_query_template_response(raw_response)

    def delete(self, name: str) -> dict:
        """Deletes the specified query template.

        Parameters
        ----------
        name : str
            The name of the query template to be deleted.

        Returns
        -------
        dict
            A dictionary containing a success message in the format of::

                {
                    "message": "string"
                }

        """
        path = f"/{self.query.tag}/{self.tag}/{name}"

        return self.client._request(method="delete", path=path)



from __future__ import annotations

import json
import os
from typing import Any, Optional, Union

import requests
from tqdm import tqdm

import thanosql._error as thanosql_error


class ThanoSQLBaseClient:
    """Base client for accessing various ThanoSQL services.

    Attributes
    ----------
    token: str
        Access token to be used in the request header.
    base_url: str
        Base URL of the ThanoSQL service.
    version: str
        Version of the API.
    url: str
        Base API URL of the ThanoSQL service that contains the base_url
        and version.

    Raises
    ------
    ThanoSQLPermissionError
        - If an invalid API token is provided.
        - If an operation is forbidden.
    ThanoSQLAlreadyExistsError
        If an object with the same name already exists.
    ThanoSQLNotFoundError
        If a requested object is not found.
    ThanoSQLValueError
        If input values are in incorrect format.
    ThanoSQLInternalError
        If an error happens while doing operations on the workspace
        or fetching data from the database.

    """

    def __init__(self, token: str, base_url: str, version: str) -> None:
        self.token: str = token
        self.base_url: str = base_url.strip("/")
        self.version: str = version

        self.url: str = f"{self.base_url}/api/{version}"

    def _create_auth_header(self) -> dict:
        return {"Authorization": f"Bearer {self.token}"}

    def _create_full_url(
        self,
        path: str = "",
        path_prefix: str = "",
        path_params: Optional[dict] = None,
        query_params: Optional[dict] = None,
    ) -> str:
        url = self.url + path

        if path_prefix:
            url = url.replace("api", f"{path_prefix}/api")

        if path_params:
            for param, value in path_params.items():
                url = url.replace(param, value)

        if query_params:
            query_params_list = []
            for param, value in query_params.items():
                query_params_list.append(f"{param}={value}")
            query_params_string = "&".join(query_params_list)
            url = f"{url}?{query_params_string}"

        return url

    def _request(
        self,
        method: str,
        path: str,
        path_prefix: str = "",
        path_params: Optional[dict] = None,
        query_params: Optional[dict] = None,
        payload: Optional[dict] = None,
        file: Optional[Union[str, os.PathLike]] = None,
        stream: bool = False,
    ) -> Any:
        full_url = self._create_full_url(
            path=path,
            path_prefix=path_prefix,
            path_params=path_params,
            query_params=query_params,
        )

        headers = self._create_auth_header()
        headers["accept"] = "application/json"

        payload_json = {}

        try:
            if file:
                payload_json["files"] = {
                    "file": (os.path.basename(file), open(file, "rb"))
                }
                if payload:
                    payload_json["files"]["body"] = (
                        None,
                        json.dumps(payload),
                        "application/json",
                    )

            elif payload is not None:
                payload_json["json"] = payload

            request_func = getattr(requests, method.lower())
            response = request_func(
                url=full_url, headers=headers, stream=stream, **payload_json
            )

            response_json = {}
            if "application/json" in response.headers.get("Content-Type", ""):
                response_json = response.json()

            if not response.ok or "error" in response_json:
                code = response.status_code
                message = "An error had occurred. Please contact ThanoSQL team."

                if "error" in response_json:
                    code = response_json["error"].get("code", code)
                    message = response_json["error"].get("message", message)

                if code in {400, 405, 422}:
                    raise thanosql_error.ThanoSQLValueError(message=message)
                elif code in {401, 403}:
                    raise thanosql_error.ThanoSQLPermissionError(message=message)
                elif code == 404:
                    raise thanosql_error.ThanoSQLNotFoundError(message=message)
                elif code == 409:
                    raise thanosql_error.ThanoSQLAlreadyExistsError(message=message)
                # includes 413 and 500, among many others
                # will show up as ThanoSQLInternalError with the message from raise_for_status
                else:
                    response.raise_for_status()

            if stream:
                filename = response.headers.get(
                    "Content-Disposition", "filename=output.bin"
                ).split("filename=")[1]
                with open(filename.strip('"'), "wb") as handle:
                    for data in tqdm(response.iter_content()):
                        handle.write(data)
                return {"message": f"Successfully downloaded {filename}."}

            if response_json:
                return response_json

            return response
        except thanosql_error.ThanoSQLError as ex:
            raise ex
        except FileNotFoundError as ex:
            raise thanosql_error.ThanoSQLNotFoundError(message=str(ex))
        except PermissionError as ex:
            raise thanosql_error.ThanoSQLPermissionError(message=str(ex))
        except requests.exceptions.JSONDecodeError as ex:
            raise thanosql_error.ThanoSQLValueError(message=str(ex))
        except TypeError as ex:
            raise thanosql_error.ThanoSQLValueError(message=str(ex))
        except Exception as e:
            raise thanosql_error.ThanoSQLInternalError(message=str(e))



from __future__ import annotations

import os
from typing import Optional

from thanosql._base_client import ThanoSQLBaseClient
from thanosql._error import ThanoSQLValueError
from thanosql.resources import (
    FileService,
    QueryService,
    SchemaService,
    TableService,
    ViewService,
)


class ThanoSQL(ThanoSQLBaseClient):
    """Client for accessing the ThanoSQL engine.

    Parameters
    ----------
    api_token: str, optional
        The workspace/engine API token. If not supplied, will be retrieved
        from the THANOSQL_API_TOKEN environment variable.
    engine_url: str, optional
        The workspace engine URL. If not supplied, will be retrieved from
        the THANOSQL_ENGINE_URL environment variable.
    api_version: str, default "v1"
        The version of the API used.

    Attributes
    ----------
    query: QueryService
        The service layer for query APIs.
    file: FileService
        The service layer for file APIs.
    schema: SchemaService
        The service layer for schema APIs.
    table: TableService
        The service layer for table APIs.
    view: ViewService
        The service layer for view APIs.

    Raises
    ------
    ThanoSQLValueError
        If API token and/or engine URL is neither passed in as parameters
        nor set as environment variables.

    """

    def __init__(
        self,
        api_token: Optional[str] = None,
        engine_url: Optional[str] = None,
        api_version: str = "v1",
    ) -> None:
        if api_token is None:
            api_token = os.environ.get("THANOSQL_API_TOKEN", "")
        if not api_token:
            raise ThanoSQLValueError(
                "Please input a valid API token. You can do this either by passing it as a parameter or setting the THANOSQL_API_TOKEN environment variable."
            )

        if engine_url is None:
            engine_url = os.environ.get("THANOSQL_ENGINE_URL", "")
        if not engine_url:
            raise ThanoSQLValueError(
                "Please input a valid engine URL. You can do this either by passing it as a parameter or setting the THANOSQL_ENGINE_URL environment variable."
            )

        super().__init__(token=api_token, base_url=engine_url, version=api_version)

    @property
    def query(self) -> QueryService:
        """Access the QueryService."""
        return QueryService(self)

    @property
    def file(self) -> FileService:
        """Access the FileService."""
        return FileService(self)

    @property
    def schema(self) -> SchemaService:
        """Access the SchemaService."""
        return SchemaService(self)

    @property
    def table(self) -> TableService:
        """Access the TableService."""
        return TableService(self)

    @property
    def view(self) -> ViewService:
        """Access the ViewService."""
        return ViewService(self)



from __future__ import annotations

import os
from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional, Union

from pydantic import TypeAdapter

from thanosql._service import ThanoSQLService
from thanosql.resources._model import BaseModel

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL


class ContentInfo(BaseModel):
    name: str
    format: Optional[str] = None
    path: str
    type: str
    size: Optional[int] = None
    writable: bool
    content: Any = None
    created_at: datetime
    updated_at: datetime


class Content(BaseModel):
    content_info: ContentInfo
    root: str


class Size(BaseModel):
    max_size: int
    used_size: int
    avail_size: int


class FileService(ThanoSQLService):
    """Service layer for file methods.

    Attributes
    ----------
    client: ThanoSQL
        The ThanoSQL client used to make requests to the engine.

    """

    def __init__(self, client: ThanoSQL) -> None:
        super().__init__(client=client, tag="fm")

    def _parse_file_content_response(self, raw_response: dict):
        file_content_adapter = TypeAdapter(Content)
        parsed_response = file_content_adapter.validate_python(raw_response)
        return parsed_response

    def get(
        self,
        path: Optional[Union[str, os.PathLike]] = None,
        option: Optional[str] = None,
    ) -> Optional[Content]:
        """Details the information of a file or directory in the specified path.

        Parameters
        ----------
        path: str or path_like
            The path to the file/directory relative to the user data root
            (default value is "/", the user data root directory itself)
        option: controls the behavior of the API
            - default (None): retrieves file/directory information
            - download: downloads a file (directory download is not possible)

        Returns
        -------
        Depending on option:
        - default (None): Returns a Content Pydantic object containing
            information on the target path
        - download: Downloads the requested file and returns nothing

        Raises
        ------
        ThanoSQLValueError
            If user attempts to download a directory (when option="download")

        """
        path_prefix = "fm"
        api_path = "/contents/"
        if path:
            api_path = api_path + str(path)
        query_params = self._create_input_dict(option=option)

        if option == "download":
            self.client._request(
                method="get",
                path=api_path,
                path_prefix=path_prefix,
                query_params=query_params,
                stream=True,
            )
        else:
            res = self.client._request(
                method="get",
                path=api_path,
                path_prefix=path_prefix,
                query_params=query_params,
            )
            return self._parse_file_content_response(res)

    def create(
        self,
        path: Optional[Union[str, os.PathLike]] = None,
        file: Optional[Union[str, os.PathLike]] = None,
    ) -> Content:
        """Uploads a file to the workspace or creates an empty folder.

        Parameters
        ----------
        path: str or path_like
            The destination path of the upload/folder creation relative to the user data root.
            Should point to a (would-be) directory (default value is "/", the user data root directory itself)
        file : str or PathLike, optional
            The file to be uploaded. If left empty, creates an empty folder in the given path.

        Returns
        -------
        A Content Pydantic object containing information on the created file/folder.

        Raises
        ------
        ThanoSQLValueError
            If path does not point to a (would-be) directory

        """
        path_prefix = "fm"
        api_path = "/contents/"
        if path:
            api_path = api_path + str(path)

        res = self.client._request(
            method="post", path=api_path, path_prefix=path_prefix, file=file
        )

        return self._parse_file_content_response(res)

    def delete(
        self,
        path: Union[str, os.PathLike],
    ) -> None:
        """Deletes the specified file or directory permanently from the workspace.

        Parameters
        ----------
        path: str or path_like
            The path to the file or directory to be removed from the workspace.
            Relative to the user data root, no default value.

        Returns
        -------
        No content

        Raises
        ------

        """
        path_prefix = "fm"
        api_path = f"/contents/{str(path)}"

        return self.client._request(
            method="delete", path=api_path, path_prefix=path_prefix
        )



from __future__ import annotations

import enum
import os
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Union

import pandas as pd
from numpy import nan
from pydantic import Field, TypeAdapter

from thanosql._error import ThanoSQLValueError
from thanosql._service import ThanoSQLService
from thanosql.resources._model import BaseModel
from thanosql.resources._record import Records

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL


class BaseColumn(BaseModel):
    default: Optional[str] = None
    is_nullable: Optional[bool] = True
    type: str
    name: str


class Column(BaseModel):
    id: Optional[int] = None
    default: Optional[str] = None
    is_nullable: Optional[bool] = True
    type: str
    name: str


class Unique(BaseModel):
    name: Optional[str] = None
    columns: Optional[List[str]] = []


class PrimaryKey(BaseModel):
    name: Optional[str] = None
    columns: Optional[List[str]] = []


class ForeignKey(BaseModel):
    name: Optional[str] = None
    reference_schema: str = "public"
    reference_column: str
    reference_table: str
    column: str


class Constraints(BaseModel):
    unique: Optional[List[Unique]] = None
    primary_key: Optional[PrimaryKey] = None
    foreign_keys: Optional[List[ForeignKey]] = None


class BaseTable(BaseModel):
    name: Optional[str] = None
    table_schema: Optional[str] = Field(alias="schema", default=None)
    columns: Optional[List[BaseColumn]] = None
    constraints: Optional[Constraints] = None


class TableObject(BaseModel):
    columns: Optional[List[BaseColumn]] = None
    constraints: Optional[Constraints] = None


class IfExists(enum.Enum):
    FAIL = "fail"
    APPEND = "append"
    REPLACE = "replace"


class TableService(ThanoSQLService):
    """Service layer for table methods.

    Attributes
    ----------
    client: ThanoSQL
        The ThanoSQL client used to make requests to the engine.
    template: TableTemplateService
        The table template service layer to access methods involving
        table templates.

    """

    def __init__(self, client: ThanoSQL) -> None:
        super().__init__(client=client, tag="table")

        self.template: TableTemplateService = TableTemplateService(client)

    def _parse_table_response(self, raw_response: dict) -> Table:
        table_adapter = TypeAdapter(Table)
        parsed_response = table_adapter.validate_python(raw_response["table"])
        parsed_response.service = self
        return parsed_response

    def list(
        self,
        schema: Optional[str] = None,
        verbose: Optional[bool] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List[Table]:
        """Lists tables stored in the workspace.

        Parameters
        ----------
        schema : str, optional
            The schema where the listed tables should reside in. If not set,
            all tables from all schemas will be included.
        verbose : bool, optional
            Whether to include the table columns and constraints in the results.
            By default, or if set to False, only retrieves the names and schemas
            of stored tables.
        offset : int, optional
            When set to n, skips the first n results and excludes them from
            the output list. Otherwise, starts the list from the first result
            stored. Must be greater than 0.
        limit : int, optional
            When set to n, limits the number of results listed to n. Otherwise,
            lists up to 100 results per call. Must range between 0 to 100.

        Returns
        -------
        List[Table]
            A list of Table objects.

        Raises
        ------
        ThanoSQLValueError
            If offset is less than 0 or if limit is not between 0 to 100 (inclusive).

        """
        path = f"/{self.tag}/"
        query_params = self._create_input_dict(
            schema=schema, verbose=verbose, offset=offset, limit=limit
        )

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        tables_adapter = TypeAdapter(List[Table])
        parsed_response = tables_adapter.validate_python(raw_response["tables"])
        for table in parsed_response:
            table.service = self
        return parsed_response

    def get(self, name: str, schema: Optional[str] = None) -> Table:
        """Shows the details of the specified table.

        Parameters
        ----------
        name : str
            The name of the table to be retrieved.
        schema : str, optional
            The schema where the table to be retrieved is in. If not
            specified, this method will look for the table in "public".

        Returns
        -------
        Table
            A Table object.

        """
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        return self._parse_table_response(raw_response)

    def update(
        self, name: str, schema: Optional[str] = None, table: Optional[BaseTable] = None
    ) -> Table:
        """Updates the specified table.

        Parameters
        ----------
        name : str
            The name of the table to be updated.
        schema : str, optional
            The schema where the table to be updated is in. If not
            specified, this method will look for the table in "public".
        table : BaseTable, optional
            BaseTable object containing changed details of the table to be
            updated. Any attribute of BaseTable can be modified, and if left
            unset, the current value will be maintained after update.
            The attributes are as follows:
            - name: new name to rename the table to
            - schema: new schema to move the table to
            - columns: new columns of the updated table. All new columns
                must be in this object, including columns that already exist
                in the original table. If this attribute is set but some
                original columns are not included, they will be removed from
                the table.
            - constraints: new constraints of the updated table. All new
                constraints must be in this object, including constraints
                that already exist in the original table. If this attribute is
                set but some original constraints are not included, they will
                be removed from the table.

        Returns
        -------
        Table
            Table object of the new table after update.

        Raises
        ------
        ThanoSQLValueError
            If the table object contains invalid formatting.

        """
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)
        payload = self._create_input_dict(table=table)

        raw_response = self.client._request(
            method="put", path=path, query_params=query_params, payload=payload
        )

        return self._parse_table_response(raw_response)

    def create(
        self,
        name: str,
        table: TableObject,
        schema: Optional[str] = None,
    ) -> Table:
        """Creates a new table.

        Parameters
        ----------
        name : str
            The name of the table to be created.
        table : TableObject
            TableObject containing the columns and constraints of the table
            to be created. In order to create an empty table, pass in an empty
            object (TableObject()).
        schema : str, optional
            The schema to save the created table in. If not specified, the table
            will be saved to "public".

        Returns
        -------
        Table
            Table object of the created table.

        Raises
        ------
        ThanoSQLValueError
            If the table object contains invalid formatting.

        """
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)
        payload = self._create_input_dict(table=table)

        raw_response = self.client._request(
            method="post", path=path, query_params=query_params, payload=payload
        )

        return self._parse_table_response(raw_response)

    def upload(
        self,
        name: str,
        file: Optional[Union[str, os.PathLike]] = None,
        df: Optional[pd.DataFrame] = None,
        schema: Optional[str] = None,
        table: Optional[TableObject] = None,
        if_exists: str = "fail",
    ) -> Table:
        """Uploads the contents of a CSV or Excel-like file or Pandas DataFrame
        into the specified table.

        Either a CSV or Excel-like (.xls, .xlsx, .xlsm, .xlsb, .odf, .ods, .odt)
        file or DataFrame must be specified. However, both should not be used
        at the same time.

        Note: If you are working with a Strict Open XML Spreadsheet (*.xlsx) file,
        change it to Excel Workbook (*.xlsx), save it, and try using it again.

        Parameters
        ----------
        name : str
            The name of the table created from the file or DataFrame.
        file : str or PathLike, optional
            CSV or Excel-like file containing tabulated data to be uploaded
            to the specified table.
        df : DataFrame, optional
            Pandas DataFrame containing data to be uploaded to the specified
            table.
        schema : str, optional
            The schema to save the created table in. If not specified, the table
            will be saved to "public".
        table : TableObject, optional
            TableObject containing the columns and constraints of the table
            to be created. If specified, the created table will follow the object
            format and no type inference is conducted. Otherwise, type
            inference will be performed and the table will be created to match
            the columns from source.
        if_exists : str, default "fail"
            What to do if table of the same name already exists. There are only
            three available values:
            - fail: fails (throws an error) if the same table exists
            - append: appends records into an existing table (columns must match
                in order to not make an error)
            - replace: deletes existing table and creates a new one with the
                given name

        Returns
        -------
        Table
            Table object of the uploaded table.

        Raises
        ------
        ThanoSQLValueError
            - If if_exists is not one of "fail", "append", or "replace".
            - If neither file nor df is used, or if both are used at the same time.
            - If file is used but it is neither CSV nor Excel-like.
            - If the file or df contains badly-formatted contents.
            - If a table body is specified but does not match the contents of the \
                file or df.
            - If if_exists is set to "append" but the new contents does not match \
                the format of the existing table.

        """
        try:
            if_exists_enum = IfExists(if_exists)
        except Exception as e:
            raise ThanoSQLValueError(str(e))

        if file and df is not None:
            raise ThanoSQLValueError(
                "Cannot use both file and DataFrame for upload at the same time."
            )

        if file:
            path = f"/{self.tag}/{name}/upload/"

            file_extension = Path(file).suffix.lower()
            if file_extension == ".csv":
                path = path + "csv"
            elif file_extension in [
                ".xls",
                ".xlsx",
                ".xlsm",
                ".xlsb",
                ".odf",
                ".ods",
                ".odt",
            ]:
                path = path + "excel"
            else:
                raise ThanoSQLValueError(
                    "Invalid format: only CSV and Excel files possible."
                )

            query_params = self._create_input_dict(
                schema=schema, if_exists=if_exists_enum.value
            )
            payload = self._create_input_dict(table=table)

            raw_response = self.client._request(
                method="post",
                path=path,
                query_params=query_params,
                payload=payload,
                file=file,
            )

            return self._parse_table_response(raw_response)

        elif df is not None:
            path = f"/{self.tag}/{name}/upload/json"

            df = df.replace({nan: None})
            df_json = df.to_dict(orient="records")
            query_params = self._create_input_dict(
                schema=schema, if_exists=if_exists_enum.value
            )
            payload = self._create_input_dict(table=table, data=df_json)

            raw_response = self.client._request(
                method="post",
                path=path,
                query_params=query_params,
                payload=payload,
                file=file,
            )

            return self._parse_table_response(raw_response)

        else:
            raise ThanoSQLValueError("No file or DataFrame provided for upload")

    def delete(self, name: str, schema: Optional[str] = None) -> dict:
        """Deletes the specified table.

        Parameters
        ----------
        name : str
            The name of the table to be deleted.
        schema : str, optional
            The schema where the table to be deleted is in. If not specified,
            this method will look for the table in "public".

        Returns
        -------
        dict
            A dictionary containing a success message, table name, and schema
            in the format of::

                {
                    "message": "string",
                    "table_name": "string",
                    "schema": "string"
                }

        """
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)

        return self.client._request(
            method="delete", path=path, query_params=query_params
        )


class Table(BaseTable):
    """Extends the BaseTable class, which has name, schema,
    columns, and constraints as attributes, with a table service
    layer to allow connection to the ThanoSQL engine.

    """

    service: Optional[TableService] = None
    """The table service layer to access the ThanoSQL client."""

    def get_records(
        self,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Records:
        """Lists the records of the table.

        Parameters
        ----------
        offset : int, optional
            When set to n, skips the first n results and excludes them from
            the output list. Otherwise, starts the list from the first result
            stored. Must be greater than 0.
        limit : int, optional
            When set to n, limits the number of results listed to n. Otherwise,
            lists up to 100 results per call. Must range between 0 to 100.

        Returns
        -------
        Records
            A Records object.

        Raises
        ------
        ThanoSQLValueError
            If offset is less than 0 or if limit is not between 0 to 100 (inclusive).

        """
        path = f"/{self.service.tag}/{self.name}/records"

        query_params = self.service._create_input_dict(
            schema=self.table_schema,
            offset=offset,
            limit=limit,
        )

        res = self.service.client._request(
            method="get",
            path=path,
            query_params=query_params,
        )
        return Records(data=res["records"], total=res["total"])

    def get_records_as_csv(
        self,
        timezone_offset: Optional[int] = None,
    ) -> None:
        """Downloads the records of the table as a CSV file.

        Parameters
        ----------
        timezone_offset : int, optional
            Timezone offset from Coordinated Universal Time (UTC).
            If not set, this value is 9, following the timezone in Seoul.
            This value is used to determine the time used in the file name.

        """
        path = f"/{self.service.tag}/{self.name}/records/csv"

        query_params = self.service._create_input_dict(
            schema=self.table_schema,
            timezone_offset=timezone_offset,
        )

        self.service.client._request(
            method="get", path=path, query_params=query_params, stream=True
        )

    def insert(
        self,
        records: List[dict],
    ) -> Table:
        """Inserts records to the specified table.

        Parameters
        ----------
        records : list of dict
            The records to be inserted in the format of a list of
            column-value pairs.

        Returns
        -------
        Table
            A Table object.

        Raises
        ------
        ThanoSQLValueError
            If the records are in an invalid format or contain invalid contents.

        """
        path = f"/{self.service.tag}/{self.name}/records"
        query_params = self.service._create_input_dict(schema=self.table_schema)

        raw_response = self.service.client._request(
            method="post", path=path, query_params=query_params, payload=records
        )

        return self.service._parse_table_response(raw_response)


class TableTemplate(BaseModel):
    name: str
    table_template: TableObject
    version: Optional[str]
    compatibility: Optional[str]
    created_at: Optional[datetime]


class TableTemplateService(ThanoSQLService):
    """Service layer for table template methods.

    Attributes
    ----------
    client: ThanoSQL
        The ThanoSQL client used to make requests to the engine.

    """

    def __init__(self, client: ThanoSQL) -> None:
        super().__init__(client=client, tag="table_template")

    def list(
        self,
        search: Optional[str] = None,
        order_by: Optional[str] = None,
        latest: Optional[bool] = None,
    ) -> List[TableTemplate]:
        """Lists table templates in the workspace.

        Parameters
        ----------
        search : str, optional
            Search keywords that the table template names in the results must
            contain. If not set, all table templates are returned by default.
        order_by : str, optional
            How to order the results. There are only three possible values:
            - recent: based on the date of creation, from most recent to oldest
            - name_asc: based on the name of the template, from A to Z
            - name_desc: based on the name of the template, from Z to A
        latest : bool, optional
            Whether to return only the latest version of each table template.
            By default, or if set to False, all versions of table templates are
            included in the results.

        Returns
        -------
        List[TableTemplate]
            A list of TableTemplate objects.

        Raises
        ------
        ThanoSQLValueError
            If order_by is not one of "recent", "name_asc", or "name_desc".

        """
        path = f"/{self.tag}/"
        query_params = self._create_input_dict(
            search=search,
            order_by=order_by,
            latest=latest,
        )

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        table_templates_adapter = TypeAdapter(List[TableTemplate])
        parsed_response = table_templates_adapter.validate_python(
            raw_response["table_templates"]
        )
        return parsed_response

    def get(self, name: str, version: Optional[str] = None) -> dict:
        """Shows the details of the specified table template.

        Parameters
        ----------
        name : str
            The name of the table template to be retrieved.
        version : str, optional
            The version of the table template to be retrieved. The value
            can either be a specific version such as "1.0", or "latest".
            If "latest" is specified, only the latest version of the table
            template will be shown. If version is not set, all versions
            will be shown.

        Returns
        -------
        dict
            A dictionary of matching table template(s) in the format of::

                {
                    "table_templates": ["TableTemplate"],
                    "versions": ["string"]
                }

        """
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(version=version)

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        table_templates_adapter = TypeAdapter(List[TableTemplate])
        parsed_response = {}
        parsed_response["table_templates"] = table_templates_adapter.validate_python(
            raw_response["table_templates"]
        )
        parsed_response["versions"] = raw_response["versions"]

        return parsed_response

    def create(
        self,
        name: str,
        table_template: TableObject,
        version: Optional[str] = None,
        compatibility: Optional[str] = None,
    ) -> TableTemplate:
        """Creates a new table template.

        Parameters
        ----------
        name : str
            The name of the table template to be created.
        table_template : TableObject
            TableObject containing the columns and constraints of the table
            template to be created. In order to create an empty table template,
            pass in an empty object (TableObject()).
        version : str, optional
            The version of the table template to be created. It must be in the
            format of "[1-9].[0-9]". If not set, it will default to "1.0".
        compatibility : str, optional
            The compatibility setting of the table template to be created. If not
            set, it will default to "ignore" (no compatibility checks).

        Returns
        -------
        TableTemplate
            TableTemplate object of the created table template.

        Raises
        ------
        ThanoSQLValueError
            - If the template name contains invalid characters or is too long.
            - If version is specified but is not in the right format.
            - If the table template contains formatting errors.

        """
        path = f"/{self.tag}/{name}"
        payload = self._create_input_dict(
            table_template=vars(table_template),
            version=version,
            compatibility=compatibility,
        )

        raw_response = self.client._request(method="post", path=path, payload=payload)

        table_template_adapter = TypeAdapter(TableTemplate)
        parsed_response = table_template_adapter.validate_python(
            raw_response["table_template"]
        )
        return parsed_response

    def delete(self, name: str, version: Optional[str] = None) -> dict:
        """Deletes the specified table template.

        Parameters
        ----------
        name : str
            The name of the table template to be removed.
        version : str, optional
            The version of the table template to be removed. If not specified, all
            versions of the table template will be removed.

        Returns
        -------
        dict
            A dictionary containing a success message and the name of the table
            template in the format of::

                {
                    "message": "string",
                    "table_template_name": "string"
                }

        """
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(version=version)

        return self.client._request(
            method="delete", path=path, query_params=query_params
        )
